import os, sys

class FakeSysExecutable(object):

  def __init__(self, python):
    self.executable = python

  def __getattr__(self, attr):
    return getattr(sys, attr)

def get_distributions():
  from pkg_resources import get_distribution
  distributions = ['setuptools', 'zc.buildout']
  try:
    import slapos.libnetworkcache
  except ImportError:
    pass
  else:
    distributions.append('slapos.libnetworkcache')
  return map(get_distribution, distributions)

def setup_script(path, python=sys.executable):
  from zc.buildout import easy_install
  try:
    if sys.executable != python:
      easy_install.sys = FakeSysExecutable(python)
    easy_install.scripts(
      ((os.path.realpath(path), 'zc.buildout.buildout', 'main'),),
      get_distributions(),
      python)
  finally:
    easy_install.sys = sys

def main():
  import shutil, subprocess, tempfile, zipfile
  eggs_dir = sys.argv[2]
  cache = sys.argv[3]
  # Install setuptools.
  src = os.path.join(cache, sys.argv[4].replace('==', '-') + '.zip')
  tmp = tempfile.mkdtemp()
  try:
    with zipfile.ZipFile(src) as zip_file:
      zip_file.extractall(tmp)
    src, = os.listdir(tmp)
    subprocess.check_call((sys.executable, 'setup.py', '-q', 'bdist_egg',
                          '--dist-dir', tmp), cwd=os.path.join(tmp, src))
    egg = os.listdir(tmp)
    egg.remove(src)
    egg, = egg
    dst = os.path.join(eggs_dir, egg)
    os.path.exists(dst) or shutil.move(os.path.join(tmp, egg), dst)
  finally:
    shutil.rmtree(tmp)
  sys.path.insert(0, dst)
  # Install other requirements given on command line.
  from pkg_resources import working_set, require
  from setuptools.command import easy_install
  reqs = sys.argv[5:]
  easy_install.main(['-mZqNxd', eggs_dir, '-f', cache] + reqs)
  working_set.add_entry(eggs_dir)
  for req in reqs:
    require(req)
  # Generate bin/buildout-rebootstrap script.
  setup_script(sys.argv[1])

if __name__ == '__main__':
  sys.exit(main())



##############################################################################
#
# Copyright (c) 2010 ViFiB SARL and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from zope.testing import renormalizing
import doctest
import pkg_resources
import re
import shutil
import sys
import unittest
import zc.buildout
import zc.buildout.testing
import zc.buildout.tests

def _setUp(cp, mkdir, write, buildout, sample_buildout, **kw):
  cp(buildout, buildout + '-orig')
  mkdir(sample_buildout, 'recipes')
  write(sample_buildout, 'recipes', 'setup.py', """
from setuptools import setup

setup(
  name = "recipes",
  entry_points = {'zc.buildout':[
    'pyinstall = pyinstall:Pyinstall',
    'pyshow = pyshow:Pyshow',
  ]},
  )
""")
  write(sample_buildout, 'recipes', 'README.txt', " ")
  write(sample_buildout, 'recipes', 'pyinstall.py', """
import os, zc.buildout, shutil, sys

class Pyinstall:

  def __init__(self, buildout, name, options):
    self.options = options
    options['executable'] = os.path.join(buildout['buildout'][
      'parts-directory'], name, 'bin', 'python')

  def install(self):
    python = self.options['executable']
    if not os.path.exists(python):
      d = os.path.dirname(python)
      os.path.exists(d) or os.makedirs(d)
      shutil.copy(sys.executable, python)
    return []

  update = install
""")
  write(sample_buildout, 'recipes', 'pyshow.py', """
import os, zc.buildout, shutil, sys

class Pyshow:

  def __init__(self, buildout, name, options):
    pass

  def install(self):
    print 'Running with:', sys.executable
    return []

  update = install
""")
  write(sample_buildout, 'buildout.cfg', """
[buildout]
develop = recipes
parts =
""")

def setUp(test):
  zc.buildout.testing.buildoutSetUp(test)
  zc.buildout.testing.install_develop('slapos.rebootstrap', test)
  test.globs['cp'] = shutil.copy
  _setUp(**test.globs)

def test_suite():
  # Note: Doctests are used, as this is the good way to test zc.buildout based
  #       applications. And zc.buildout.testing.buildoutSetUp does *NOT* support
  #       non-doctest suites
  kwargs = dict(setUp=setUp,
      tearDown=zc.buildout.testing.buildoutTearDown,
      checker=renormalizing.RENormalizing([
                        (re.compile('--prefix=\S+sample-buildout'),
                         '--prefix=/sample_buildout'),
                        (re.compile('\s/\S+sample-buildout'),
                         ' /sample_buildout'),
                        (re.compile(sys.executable),
                         '/system_python'),
                        zc.buildout.testing.normalize_path,
                        zc.buildout.testing.not_found,
                        ]),
    )
  test_list = []
  for text in pkg_resources.resource_listdir(__name__, '.'):
    if text.endswith('.txt'):
      test_list.append(doctest.DocFileSuite(text, **kwargs))
  suite = unittest.TestSuite(test_list)
  return suite



##############################################################################
#
# Copyright (c) 2010-2017 ViFiB SARL and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import logging, os, shutil, subprocess, sys, tempfile
from zc.buildout import easy_install, UserError


class extension(object):

  def __init__(self, buildout):
    self.buildout = buildout
    # fetch section to build python (default value is 'buildout')
    self.python_section = buildout['buildout']['python'].strip()
    self.wanted_python = buildout[self.python_section]['executable']
    if sys.executable != self.wanted_python:
      self.hook('_setup_directories')
    elif buildout._parts:
      self._frozen = frozenset(buildout._parts)
      self.hook('_compute_part_signatures')

  def hook(self, attr):
    buildout = self.buildout
    getattr(buildout, attr)
    def wrapper(*args, **kw):
      delattr(buildout, attr)
      return getattr(self, attr)(*args, **kw)
    setattr(buildout, attr, wrapper)

  def _setup_directories(self):
    logger = logging.getLogger(__name__)
    buildout = self.buildout
    logger.info(
      "Make sure that the section %r won't be reinstalled after rebootstrap."
      % self.python_section)
    # We hooked in such a way that all extensions are loaded. Do not reload.
    buildout._load_extensions
    buildout._load_extensions = lambda: None
    # workaround for the install command,
    # which ignores dependencies when parts are specified
    # (the only sections we have accessed so far are those that are required
    #  to build the wanted Python)
    buildout.install(buildout._parts) # [self.python_section]

    logger.info("""
************ REBOOTSTRAP: IMPORTANT NOTICE ************
bin/buildout is being reinstalled right now, as new python:
  %(wanted_python)s
is available, and buildout is using another python:
  %(running_python)s
Buildout will be restarted automatically to have this change applied.
************ REBOOTSTRAP: IMPORTANT NOTICE ************
""" % dict(wanted_python=self.wanted_python, running_python=sys.executable))

    installed = sys.argv[0]
    new_bin = installed + '-rebootstrap'
    if not os.path.exists(new_bin):
      from .bootstrap import get_distributions, setup_script
      if subprocess.call((self.wanted_python, '-c',
          'import sys; sys.exit(sys.version_info[:2] == %r)'
          % (sys.version_info[:2],))):
        setup_script(new_bin, self.wanted_python)
      else:
        # With a different version of Python,
        # we must reinstall required eggs from source.
        from pkg_resources import resource_string
        with Cache(buildout['buildout']) as cache:
          subprocess.check_call([self.wanted_python, '-c',
            resource_string(__name__, 'bootstrap.py'),
            new_bin, cache._dest, cache.tmp,
            ] + list(map(cache, get_distributions())))

    shutil.copy(new_bin, installed)
    os.execv(self.wanted_python, [self.wanted_python] + sys.argv)

  def _compute_part_signatures(self, install_parts):
    # Skip signature check for parts that were required to build the wanted
    # Python. Signatures differ when switching to a different version.
    buildout = self.buildout
    buildout._compute_part_signatures(install_parts)
    installed_part_options = buildout.installed_part_options
    for part in self._frozen.intersection(install_parts):
      buildout[part]['__buildout_signature__'] = \
        installed_part_options[part]['__buildout_signature__']


class Cache(easy_install.Installer):

  def __init__(self, buildout):
    easy_install.Installer.__init__(self,
      buildout['eggs-directory'],
      buildout.get('find-links', '').split())

  def __enter__(self):
    self.tmp = self._download_cache or tempfile.mkdtemp('get_dist')
    return self

  def __exit__(self, t, v, tb):
    if self.tmp is not self._download_cache:
      shutil.rmtree(self.tmp)
    del self.tmp

  def __call__(self, dist):
    req = dist.as_requirement()
    cache = self._download_cache
    if cache:
      from pkg_resources import SOURCE_DIST
      for avail in self._index[dist.project_name]:
        if (avail.version == dist.version and
            avail.precedence == SOURCE_DIST and
            cache == os.path.dirname(avail.location)):
          return str(req)
    avail = self._obtain(req, True)
    if avail is None:
      raise UserError("Couldn't find a distribution for %r" % str(req))
    if self._fetch(avail, self.tmp, cache) is None:
      raise UserError("Couldn't download distribution %s." % avail)
    return str(req)
